import os
import time
import json
import logging
import threading
import numpy as np
import multiprocessing as mp
from ..utils import save, load, AutoGluonEarlyStop
import distributed
from distributed import Queue, Variable
from distributed.comm.core import CommClosedError

logger = logging.getLogger(__name__)

__all__ = ['DistStatusReporter',
           'FakeReporter',
           'DistSemaphore',
           'Communicator',
           'LocalStatusReporter']


class FakeReporter(object):
    """FakeReporter for internal use in final fit
    """
    def __call__(self, **kwargs):
        pass


class DistStatusReporter(object):
    """Report status through the training scheduler.

    Example
    -------
    >>> @autogluon_method
    >>> def train_func(config, reporter):
    ...     reporter(accuracy=0.1)
    """

    def __init__(self, remote=None):
        self._queue = Queue(client=remote)
        self._stop = Variable(client=remote)
        self._stop.set(False)
        self._continue_semaphore = DistSemaphore(0, remote)

    def __call__(self, **kwargs):
        """Report updated training status.
        Pass in `done=True` when the training job is completed.

        Args:
            kwargs: Latest training result status.

        Example
        _______
        >>> reporter(accuracy=1, training_iters=4)
        """
        logger.debug('Reporting {}'.format(json.dumps(kwargs)))
        try:
            self._queue.put(kwargs.copy())
        except RuntimeError:
            return
        self._continue_semaphore.acquire()
        if self._stop.get():
            raise AutoGluonEarlyStop('Stopping!')

    def fetch(self, block=True):
        try:
            kwargs = self._queue.get()
        except CommClosedError:
            return {}
        return kwargs

    def terminate(self):
        self._stop.set(True)
        self._continue_semaphore.release()

    def move_on(self):
        self._continue_semaphore.release()

    def save_dict(self, **state_dict):
        raise NotImplementedError

    def get_dict(self):
        raise NotImplementedError

    def __repr__(self):
        reprstr = self.__class__.__name__
        return reprstr


class MODistStatusReporter(DistStatusReporter):
    """Report status through the training of a multi-objective scheduler.

    Example
    -------
    >>> @autogluon_method
    >>> def train_func(config, reporter):
    ...     reporter(accuracy=1, f_score=1, training_iters=4)
    """
    def __init__(self, objectives, weights, scalarization_opts, remote=None):
        super().__init__(remote)
        self.objectives = objectives
        self.weights = weights
        self.scalarization_options = scalarization_opts
    
    def __call__(self, **kwargs):
        """Report updated training status.
        Pass in `done=True` when the training job is completed.

        Args:
            kwargs: Latest training result status. Reporter requires access to
            all objectives of interest.

        Example
        _______
        >>> reporter(accuracy=1, f_score=1, training_iters=4)
        """
        try:
            v = np.array([kwargs[k] for k in self.objectives])
        except KeyError:
            raise KeyError("Reporter requires accesss to all objective values.\
                Please ensure you return all required values.")

        if self.scalarization_options["algorithm"] == "random_weights":
            scalarization = max([w @ v for w in self.weights])
        elif self.scalarization_options["algorithm"] == "parego":
            rho = self.scalarization_options["rho"]
            scalarization = [max(w * v) + rho * (w @ v) for w in self.weights]
            scalarization = max(scalarization)
        elif self.scalarization_options["algorithm"] == "golovin":
            scalarization = [np.min(np.clip(v / w, a_min=0, a_max=None)) ** len(w) for w in self.weights]
            scalarization = max(scalarization)
        else:
            raise ValueError("Specified scalarization algorithm is unknown. \
                Valid algorithms are 'random_weights' and 'parego'.")
        kwargs["_SCALARIZATION"] = scalarization

        super().__call__(**kwargs)


class LocalStatusReporter(object):
    """Local status reporter (automatically created by communicator)
    Example
    -------
    >>> def train_func(config, reporter):
    ...     assert isinstance(reporter, StatusReporter)
    ...     reporter(timesteps_this_iter=1)
    """

    def __init__(self, dict_path=None):#, result_queue, continue_semaphore):
        self._queue = mp.Queue(1)
        self._stop = mp.Value('i', 0)
        self._continue_semaphore = mp.Semaphore(0)
        self._save_dict = False
        self.dict_path = dict_path

    def __call__(self, **kwargs):
        """Report updated training status.
        Pass in `done=True` when the training job is completed.

        Args:
            kwargs: Latest training result status.
        Example
        -------
        >>> reporter(accuracy=1, training_iters=4)
        """
        self._queue.put(kwargs.copy(), block=True)
        logger.debug('StatusReporter reporting: {}'.format(json.dumps(kwargs)))

        self._continue_semaphore.acquire()
        if self._stop.value:
            raise AutoGluonEarlyStop

    def fetch(self, block=True):
        kwargs = self._queue.get(block=block)
        return kwargs

    def move_on(self):
        self._continue_semaphore.release()

    def terminate(self):
        self._stop.value = 1
        self._continue_semaphore.release()

    def save_dict(self, **state_dict):
        """Save the serializable state_dict
        """
        logger.debug('Saving the task dict to {}'.format(self.dict_path))
        save(state_dict, self.dict_path)

    def has_dict(self):
        logger.debug('has_dict {}'.format(os.path.isfile(self.dict_path)))
        return os.path.isfile(self.dict_path)

    def get_dict(self):
        return load(self.dict_path)

    def __repr__(self):
        reprstr = self.__class__.__name__
        return reprstr


class Communicator(threading.Thread):
    def __init__(self, process, local_reporter, dist_reporter):
        super().__init__()
        self.process = process
        self.local_reporter = local_reporter
        self.dist_reporter = dist_reporter
        self._stop_event = threading.Event()

    def run(self):
        while self.process.is_alive():

            # breaking communication if process raises exception
            if self.process.exception is not None:
                error, traceback = self.process.exception
                self.local_reporter.terminate()
                self.dist_reporter(done=True, traceback=traceback)
                self.process.join()
                break

            try:
                # waiting until process reports results or raises exception
                if self.local_reporter._queue.empty():
                    continue
                reported_result = self.local_reporter.fetch()
            except BrokenPipeError:
                break

            try:
                self.dist_reporter(**reported_result)
                self.local_reporter.move_on()
            except AutoGluonEarlyStop:
                self.local_reporter.terminate()

            if reported_result.get('done', False):
                self.process.join()
                break

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    @classmethod
    def Create(cls, process, local_reporter, dist_reporter):
        communicator = cls(process, local_reporter, dist_reporter)
        communicator.start()
        return communicator

    def __repr__(self):
        reprstr = self.__class__.__name__
        return reprstr


class DistSemaphore(object):
    def __init__(self, value, remote=None):
        self._queue = Queue(client=remote)
        for i in range(value):
            self._queue.put(1)

    def acquire(self):
        try:
            _ = self._queue.get()
        except distributed.comm.core.CommClosedError:
            pass

    def release(self):
        self._queue.put(1)

    def __repr__(self):
        reprstr = self.__class__.__name__
        return reprstr



import autograd.numpy as anp

from .base import KernelFunction

__all__ = ['ProductKernelFunction']


class ProductKernelFunction(KernelFunction):
    """
    Given two kernel functions K1, K2, this class represents the product kernel
    function given by

        ((x1, x2), (y1, y2)) -> K(x1, y1) * K(x2, y2)

    We assume that parameters of K1 and K2 are disjoint.
    """
    def __init__(self, kernel1: KernelFunction, kernel2: KernelFunction,
                 name_prefixes=None, **kwargs):
        """
        :param kernel1: Kernel function K1
        :param kernel2: Kernel function K2
        :param name_prefixes: Name prefixes for K1, K2 used in get_params
        """
        super(ProductKernelFunction, self).__init__(
            kernel1.dimension + kernel2.dimension, **kwargs)
        self.kernel1 = kernel1
        self.kernel2 = kernel2
        if name_prefixes is None:
            self.name_prefixes = ['kernel1', 'kernel2']
        else:
            assert len(name_prefixes) == 2
            self.name_prefixes = name_prefixes

    def forward(self, X1, X2):
        d1 = self.kernel1.dimension
        X1_1 = X1[:, :d1]
        X1_2 = X1[:, d1:]
        X2_1 = X2[:, :d1]
        X2_2 = X1[:, d1:]
        kmat1 = self.kernel1(X1_1, X2_1)
        kmat2 = self.kernel2(X1_2, X2_2)
        return kmat1 * kmat2

    def diagonal(self, X):
        d1 = self.kernel1.dimension
        X1 = X[:, :d1]
        X2 = X[:, d1:]
        diag1 = self.kernel1.diagonal(X1)
        diag2 = self.kernel2.diagonal(X2)
        return diag1 * diag2

    def diagonal_depends_on_X(self):
        return (self.kernel1.diagonal_depends_on_X() or
                self.kernel2.diagonal_depends_on_X())

    def param_encoding_pairs(self):
        """
        Note: We assume that K1 and K2 have disjoint parameters, otherwise
        there will be a redundancy here.
        """
        return self.kernel1.param_encoding_pairs() + \
               self.kernel2.param_encoding_pairs()

    def get_params(self):
        result = dict()
        prefs = [k + '_' for k in self.name_prefixes]
        for pref, kernel in zip(prefs, [self.kernel1, self.kernel2]):
            result.update({
                (pref + k): v for k, v in kernel.get_params().items()})
                
        return result

    def set_params(self, param_dict):
        prefs = [k + '_' for k in self.name_prefixes]
        for pref, kernel in zip(prefs, [self.kernel1, self.kernel2]):
            len_pref = len(pref)
            stripped_dict = {
                k[len_pref:]: v for k, v in param_dict.items()
                if k.startswith(pref)}
            kernel.set_params(stripped_dict)



from .import remote, resource
from .resource import get_cpu_count, get_gpu_count

# schedulers
from .scheduler import *
from .fifo import *
from .hyperband import *
from .mo_hyperband import *
from .mo_asha import *
from .rl_scheduler import *



# adapted from https://github.com/makinacorpus/easydict
class EasyDict(dict):
    def __init__(self, d=None, **kwargs):
        if d is None:
            d = {}
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            setattr(self, k, v)
        # Class attributes
        for k in self.__class__.__dict__.keys():
            if not (k.startswith('__') and k.endswith('__')) and not k in ('update', 'pop'):
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name, value):
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x)
                     if isinstance(x, dict) else x for x in value]
        elif isinstance(value, dict) and not isinstance(value, self.__class__):
            value = self.__class__(value)
        super(EasyDict, self).__setattr__(name, value)
        super(EasyDict, self).__setitem__(name, value)

    __setitem__ = __setattr__

    def update(self, e=None, **f):
        d = e or dict()
        d.update(f)
        for k in d:
            setattr(self, k, d[k])

    def pop(self, k, d=None):
        delattr(self, k)
        return super(EasyDict, self).pop(k, d)




import numpy as np

from ..utils import try_import_torch


def tune_temperature_scaling(y_val_probs: np.ndarray, y_val: np.ndarray, init_val: float = 1, max_iter: int = 1000,
                             lr: float = 0.01):
    '''
    Tunes a temperature scalar term that divides the logits produced by autogluon model. Logits are generated
    by natural log the predicted probs from model then divides by a temperature scalar, which is tuned
    to minimize cross entropy on validation set.

    Parameters:
    -----------
    y_val_probs: numpy ndarray
        Predictive probabilities by model on validation set
    y_val: numpy ndarray
        The labels to the validation set
    init_val: float
        Initial value for temperature scalar term
    max_iter: int
        The maximum number of iterations to step in tuning
    lr: float
        The initial learning rate

    Return:
    float: The temperature scaling term
    '''
    try_import_torch()
    import torch

    y_val_tensor = torch.tensor(y_val)
    temperature_param = torch.nn.Parameter(torch.ones(1).fill_(init_val))
    logits = torch.tensor(np.log(y_val_probs))
    nll_criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.LBFGS([temperature_param], lr=lr, max_iter=max_iter)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)

    def temperature_scale_step():
        optimizer.zero_grad()
        temp = temperature_param.unsqueeze(1).expand(logits.size(0), logits.size(1))
        new_logits = (logits / temp)
        loss = nll_criterion(new_logits, y_val_tensor)
        loss.backward()
        scheduler.step()
        return loss

    optimizer.step(temperature_scale_step)

    return temperature_param.item()



import logging
import numpy as np

from .hyperband_promotion import PromotionRungSystem

logger = logging.getLogger(__name__)

# exploration strategies
EPS_NET = "eps_net"
NSGA_II = "nsga_ii"


class MOPromotionRungSystem(PromotionRungSystem):
    r"""This is an adaption of the standard PromotionRungSystem class to the
    multi-objective domain. Its main functions are eps-net and nsga-ii filters.

    Arguments and handling follow the parent class. One addtional parameter is
    to specify the exploration strategy.

    Args:
        strategy : string
            Indicates the explorations strategy. Supported strategies are
            "eps_net" and "nsga_ii".
    """

    def __init__(self, rung_levels, promote_quantiles, max_t, strategy):
        assert strategy in [EPS_NET, NSGA_II], "Invalid selection strategy"
        self.strategy = strategy
        super().__init__(rung_levels, promote_quantiles, max_t)

    def _find_promotable_config(self, recorded, prom_quant, config_key=None):
        """
        Scans the top prom_quant fraction of recorded (sorted w.r.t. reward
        value) for config not yet promoted. If config_key is given, the key
        must also be equal to config_key.

        :param recorded: Dict to scan
        :param prom_quant: Quantile for promotion
        :param config_key: See above
        :return: Key of config if found, otherwise None
        """
        num_recorded = len(recorded)
        ret_key = None
        if num_recorded >= int(round(1.0 / prom_quant)):
            # Search for not yet promoted config in the top
            # prom_quant fraction
            def filter_pred(k, v):
                return (not v[1]) and (config_key is None or k == config_key)

            num_top = int(num_recorded * prom_quant)

            # generate matrix from individual reward vectors
            evaluations = np.array([v[0] for _, v in recorded.items()])

            if self.strategy == EPS_NET:
                ranked_top = get_eps_net_ranking(evaluations, num_top)
            else:  # NSGA_II
                ranked_top = get_nsga_ii_ranking(evaluations, num_top)

            items = list(recorded.items())
            top_list = [items[i] for i in ranked_top]

            try:
                ret_key = next(k for k, v in top_list if filter_pred(k, v))
            except StopIteration:
                ret_key = None
        return ret_key

    @staticmethod
    def _num_promotable_config(recorded, prom_quant):
        raise NotImplementedError()


# -----------------------------------------------------------------------------
# Filtration techniques
# -----------------------------------------------------------------------------

def fast_nondominated_sort(values: np.array, num_samples: int):
    # We assume a 2d np array of dim (n_candidates, n_objectives)
    # This functions assumes a minimization problem. Implementation is based
    # on the NSGA-II paper
    assert values.shape[0] > 0, "Need to provide at least 1 point."
    assert values.shape[0] >= num_samples, "Need at least enough points to \
                                            meet num_samples."
    domination_counts = np.zeros(values.shape[0])
    dominated_solutions = [[] for _ in range(values.shape[0])]
    fronts = [[]]
    ranks = np.zeros(values.shape[0])

    for i, v1 in enumerate(values):
        for j, v2 in enumerate(values):
            if np.alltrue(v1 < v2):  # v1 dominates v2
                dominated_solutions[i].append(j)
            elif np.alltrue(v2 < v1):  # v2 dominates v1
                domination_counts[i] += 1

        if domination_counts[i] == 0:
            ranks[i] = 1
            fronts[0].append(i)

    i = 0
    n_selected = len(fronts[0])
    while n_selected < num_samples:
        assert len(fronts[i]) > 0, "Cannot select from empty front"
        tmp = []
        for j in fronts[i]:
            for k in dominated_solutions[j]:
                domination_counts[k] -= 1
                if domination_counts[k] == 0:
                    ranks[k] = i + 1
                    tmp.append(k)
        i += 1
        n_selected += len(tmp)
        fronts.append(tmp)
    assert n_selected >= num_samples, "Could not assign enough samples"
    return ranks, fronts


def compute_eps_net(points: np.array, num_samples: int = None):
    """Sparsify a numpy matrix of points returning `num_samples` points that
    are as spread out as possible. Iteratively select the point that is the
    furthest from priorly selected points.
    :param points: numpy array of shape num_points x num_objectives
    :param num_samples: number of points to return
    :return: indices
    """
    assert points.shape[0] > 0, "Need to provide at least 1 point."

    def dist(points, x):
        return np.min([np.linalg.norm(p - x) for p in points])
    n = points.shape[0]
    eps_net = [0]
    indices_remaining = set(range(1, n))
    if num_samples is None:
        num_samples = n
    while len(eps_net) < num_samples and len(indices_remaining) > 0:
        # compute argmin dist(pts[i \not in eps_net], x)
        dist_max = -1
        best_i = 0
        for i in indices_remaining:
            cur_dist = dist(points[eps_net], points[i])
            if cur_dist > dist_max:
                best_i = i
                dist_max = cur_dist
        eps_net.append(best_i)
        indices_remaining.remove(best_i)
    return eps_net


def get_eps_net_ranking(points: np.array, num_top: int):
    """Produces sorted list containing the best indices
    :param points: Numpy array containing all previous evaluations
    :return: List of num_top indices
    """
    _, fronts = fast_nondominated_sort(points, num_top)
    ranked_ids = []

    i = 0
    n_selected = 0
    while n_selected < num_top:
        front = fronts[i]
        local_order = compute_eps_net(points[front],
                                      num_samples=(num_top - n_selected))
        ranked_ids += [front[j] for j in local_order]
        i += 1
        n_selected += len(local_order)
    assert len(ranked_ids) == num_top, "Did not assign correct number of \
                                        points to eps-net"
    return ranked_ids


def crowding_distance_assignment(front_points: np.array):
    assert front_points.shape[0] > 0, "Error no empty fronts are allowed"
    distances = np.zeros(front_points.shape[0])
    n_objectives = front_points.shape[1]

    for m in range(n_objectives):  # Iterate through objectives
        vs = [(front_points[i][m], i) for i in range(front_points.shape[0])]
        vs.sort()
        # last and first element have inf distance
        distances[vs[0][1]] = np.inf
        distances[vs[-1][1]] = np.inf
        ms = [front_points[i][m] for i in range(front_points.shape[0])]
        scale = max(ms) - min(ms)
        if scale == 0:
            scale = 1
        for j in range(1, front_points.shape[0] - 1):
            distances[vs[j][1]] += (vs[j + 1][0] - vs[j - 1][0]) / scale

    # determine local order
    dist_id = [(-distances[i], i) for i in range(front_points.shape[0])]
    dist_id.sort()
    local_order = [d[1] for d in dist_id]
    return local_order


def get_nsga_ii_ranking(points: np.array, num_top: int):
    """Produces sorted list containing the best indices
    :param points: Numpy array containing all previous evaluations
    :return: List of num_top indices
    """
    _, fronts = fast_nondominated_sort(points, num_top)
    ranked_ids = []

    i = 0
    n_selected = 0
    while n_selected < num_top:
        front = fronts[i]
        local_order = crowding_distance_assignment(points[front])
        ranked_ids += [front[j] for j in local_order]
        i += 1
        n_selected += len(local_order)
    return ranked_ids[:num_top]
